import re
import six
import time
from datetime import datetime
from ..logger import logger

true_pattern = re.compile(r"true", flags=re.IGNORECASE)

KylinType = dict(
    CHAR=six.text_type,
    VARCHAR=six.text_type,
    STRING=six.text_type,
    DECIMAL=float,
    DOUBLE=float,
    FLOAT=float,
    BIGINT=int,
    LONG=int,
    INTEGER=int,
    INT=int,
    TINYINT=int,
    SMALLINT=int,
    INT4=int,
    LONG8=int,
    BOOLEAN=lambda x: bool(re.search(true_pattern, x)),
    DATE=lambda x: datetime.strptime(x, "%Y-%m-%d").date(),
    DATETIME=lambda x: datetime.strptime(x, "%Y-%m-%d %H:%M:%S"),
    TIMESTAMP=lambda x: time.mktime(datetime.strptime(
        x, "%Y-%m-%d %H:%M:%S").timetuple()),
)


def kylin_to_python(_type, s):
    try:
        return KylinType.get(str(_type).upper())(s) if s else s
    except KeyError:
        logger.error(
            "CONVERT ERROR, type: {}, raw string: {}".format(_type, s))



from .kylinpy import Kylinpy
from .utils.kylin_types import kylin_to_python
from .logger import logger
from .errors import KylinDBAPIError, KylinConnectionError


class Cursor(object):
    def __init__(self, connection):
        self.connection = connection
        self._arraysize = 1

        self.description = None
        self.rowcount = -1
        self.results = None
        self.fetched_rows = 0

    def callproc(self):
        logger.warn('Stored procedures not supported in Kylin')

    def close(self):
        logger.debug('Cursor close called')

    def execute(self, query, *params):
        # todo query params
        resp = self.connection.query(query).get('data')
        self.description = [[
            c['label'],
            c['columnTypeName'].lower(),
            c['displaySize'],
            0,
            c['precision'],
            c['scale'],
            c['isNullable']
        ] for c in resp['columnMetas']]

        self.results = [[
            kylin_to_python(resp['columnMetas'][idx]['columnTypeName'], cell) for (idx, cell) in enumerate(row)
        ] for row in resp['results']]
        self.rowcount = len(self.results)
        self.fetched_rows = 0
        return self.rowcount

    def executemany(self, query, seq_params=[]):
        results = []
        for param in seq_params:
            self.execute(query, param)
            results.extend(self.results)

        self.results = results
        self.rowcount = len(self.results)
        self.fetched_rows = 0
        return self.rowcount

    def fetchone(self):
        if self.fetched_rows < self.rowcount:
            row = self.results[self.fetched_rows]
            self.fetched_rows += 1
            return row
        else:
            return None

    def fetchmany(self, size=None):
        fetched_rows = self.fetched_rows
        size = size or self.arraysize
        self.fetched_rows = fetched_rows + size
        return self.results[fetched_rows: self.fetched_rows]

    def fetchall(self):
        fetched_rows = self.fetched_rows
        self.fetched_rows = self.rowcount
        return self.results[fetched_rows:]

    def nextset(self):
        logger.warn('Nextset operation not supported in Kylin')

    @property
    def arraysize(self):
        return self._arraysize

    @arraysize.setter
    def arraysize(self, array_size):
        self._arraysize = array_size

    def setinputsizes(self):
        logger.warn('setinputsize not supported in Kylin')

    def setoutputsize(self):
        logger.warn('setoutputsize not supported in Kylin')


class KylinDB(Kylinpy):
    paramstyle = 'pyformat'
    threadsafety = 2
    apilevel = '2.0'
    Error = KylinDBAPIError

    def __init__(self, **kwargs):
        super(KylinDB, self).__init__(**kwargs)

    @classmethod
    def connect(cls, **kwargs):
        try:
            return cls(**kwargs)
        except TypeError:
            raise KylinConnectionError

    def close(self):
        return

    def commit(self):
        return

    def rollback(self):
        return

    def cursor(self):
        return Cursor(self)



import re

from sqlalchemy.types import (
    CHAR,
    VARCHAR,
    INTEGER,
    BIGINT,
    SMALLINT,
    FLOAT,
    DECIMAL,
    BOOLEAN,
    DATE,
    DATETIME,
    TIMESTAMP,
)

KylinType = dict(
    CHAR=CHAR,
    STRING=VARCHAR,
    VARCHAR=VARCHAR,
    DECIMAL=DECIMAL,
    DOUBLE=FLOAT,
    FLOAT=FLOAT,
    BIGINT=BIGINT,
    LONG=BIGINT,
    INTEGER=INTEGER,
    INT=INTEGER,
    TINYINT=SMALLINT,
    SMALLINT=SMALLINT,
    INT4=BIGINT,
    LONG8=BIGINT,
    BOOLEAN=BOOLEAN,
    DATE=DATE,
    DATETIME=DATETIME,
    TIMESTAMP=TIMESTAMP,
)


def kylin_to_sqla(s):
    type_re = re.compile('^({})\(?(\d+)?,?(\d+)?\)?$'.format(
        '|'.join(KylinType.keys())
    ), flags=re.IGNORECASE)
    try:
        type_tuple = type_re.match(s).groups()
        _type, _args = type_tuple[0].upper(), [int(e)
                                               for e in type_tuple[1:] if e]
    except AttributeError:
        pass

    return KylinType.get(_type)(*_args)



from __future__ import absolute_import

from .logger import logger


class _BaseError(Exception):
    def __init__(self, msg):
        logger.debug('''
==========================[RAW ERROR]===============================
 ERROR: %s \n
==========================[RAW ERROR]===============================
        ''', msg)
        self.msg = msg or ''
        super(_BaseError, self).__init__(msg)

    def __str__(self):
        return self.msg


class KylinUnauthorizedError(_BaseError):
    """Exception raised by unauthorized"""


class KylinUserDisabled(_BaseError):
    """Exception raised by unauthorized"""


class KylinConnectionError(_BaseError):
    """Exception raised by connection refused"""


class KylinError(_BaseError):
    """Exception raised by otherwise"""


class KylinConfusedResponse(_BaseError):
    """Exception raised by confused response"""


class KylinDBAPIError(_BaseError):
    """Exception raised dbapi error"""


class KAPOnlyError(_BaseError):
    def __init__(self):
        super(KAPOnlyError, self).__init__(
            'No such API in Apache Kylin, this is KAP only API')



from __future__ import absolute_import

CALCITE_KEYWORDS = set([
    'A', 'ABS', 'ABSOLUTE', 'ACTION',
    'ADA', 'ADD', 'ADMIN', 'AFTER',
    'ALL', 'ALLOCATE', 'ALLOW', 'ALTER',
    'ALWAYS', 'AND', 'ANY', 'APPLY',
    'ARE', 'ARRAY', 'ARRAY_MAX_CARDINALITY', 'AS',
    'ASC', 'ASENSITIVE', 'ASSERTION', 'ASSIGNMENT',
    'ASYMMETRIC', 'AT', 'ATOMIC', 'ATTRIBUTE',
    'ATTRIBUTES', 'AUTHORIZATION', 'AVG', 'BEFORE',
    'BEGIN', 'BEGIN_FRAME', 'BEGIN_PARTITION', 'BERNOULLI',
    'BETWEEN', 'BIGINT', 'BINARY', 'BIT',
    'BLOB', 'BOOLEAN', 'BOTH', 'BREADTH',
    'BY', 'C', 'CALL', 'CALLED',
    'CARDINALITY', 'CASCADE', 'CASCADED', 'CASE',
    'CAST', 'CATALOG', 'CATALOG_NAME', 'CEIL',
    'CEILING', 'CENTURY', 'CHAIN', 'CHAR',
    'CHARACTER', 'CHARACTERISTICS', 'CHARACTERS', 'CHARACTER_LENGTH',
    'CHARACTER_SET_CATALOG', 'CHARACTER_SET_NAME', 'CHARACTER_SET_SCHEMA', 'CHAR_LENGTH',
    'CHECK', 'CLASSIFIER', 'CLASS_ORIGIN', 'CLOB',
    'CLOSE', 'COALESCE', 'COBOL', 'COLLATE',
    'COLLATION', 'COLLATION_CATALOG', 'COLLATION_NAME', 'COLLATION_SCHEMA',
    'COLLECT', 'COLUMN', 'COLUMN_NAME', 'COMMAND_FUNCTION',
    'COMMAND_FUNCTION_CODE', 'COMMIT', 'COMMITTED', 'CONDITION',
    'CONDITION_NUMBER', 'CONNECT', 'CONNECTION', 'CONNECTION_NAME',
    'CONSTRAINT', 'CONSTRAINTS', 'CONSTRAINT_CATALOG', 'CONSTRAINT_NAME',
    'CONSTRAINT_SCHEMA', 'CONSTRUCTOR', 'CONTAINS', 'CONTINUE',
    'CONVERT', 'CORR', 'CORRESPONDING', 'COUNT',
    'COVAR_POP', 'COVAR_SAMP', 'CREATE', 'CROSS',
    'CUBE', 'CUME_DIST', 'CURRENT', 'CURRENT_CATALOG',
    'CURRENT_DATE', 'CURRENT_DEFAULT_TRANSFORM_GROUP', 'CURRENT_PATH', 'CURRENT_ROLE',
    'CURRENT_ROW', 'CURRENT_SCHEMA', 'CURRENT_TIME', 'CURRENT_TIMESTAMP',
    'CURRENT_TRANSFORM_GROUP_FOR_TYPE', 'CURRENT_USER', 'CURSOR', 'CURSOR_NAME',
    'CYCLE', 'DATA', 'DATABASE', 'DATE',
    'DATETIME_INTERVAL_CODE', 'DATETIME_INTERVAL_PRECISION', 'DAY', 'DEALLOCATE',
    'DEC', 'DECADE', 'DECIMAL', 'DECLARE',
    'DEFAULT', 'DEFAULTS', 'DEFERRABLE', 'DEFERRED',
    'DEFINE', 'DEFINED', 'DEFINER', 'DEGREE',
    'DELETE', 'DENSE_RANK', 'DEPTH', 'DEREF',
    'DERIVED', 'DESC', 'DESCRIBE', 'DESCRIPTION',
    'DESCRIPTOR', 'DETERMINISTIC', 'DIAGNOSTICS', 'DISALLOW',
    'DISCONNECT', 'DISPATCH', 'DISTINCT', 'DOMAIN',
    'DOUBLE', 'DOW', 'DOY', 'DROP',
    'DYNAMIC', 'DYNAMIC_FUNCTION', 'DYNAMIC_FUNCTION_CODE', 'EACH',
    'ELEMENT', 'ELSE', 'EMPTY', 'END',
    'END-EXEC', 'END_FRAME', 'END_PARTITION', 'EPOCH',
    'EQUALS', 'ESCAPE', 'EVERY', 'EXCEPT',
    'EXCEPTION', 'EXCLUDE', 'EXCLUDING', 'EXEC',
    'EXECUTE', 'EXISTS', 'EXP', 'EXPLAIN',
    'EXTEND', 'EXTERNAL', 'EXTRACT', 'FALSE',
    'FETCH', 'FILTER', 'FINAL', 'FIRST',
    'FIRST_VALUE', 'FLOAT', 'FLOOR', 'FOLLOWING',
    'FOR', 'FOREIGN', 'FORTRAN', 'FOUND',
    'FRAC_SECOND', 'FRAME_ROW', 'FREE', 'FROM',
    'FULL', 'FUNCTION', 'FUSION', 'G',
    'GENERAL', 'GENERATED', 'GEOMETRY', 'GET',
    'GLOBAL', 'GO', 'GOTO', 'GRANT',
    'GRANTED', 'GROUP', 'GROUPING', 'GROUPS',
    'HAVING', 'HIERARCHY', 'HOLD', 'HOUR',
    'IDENTITY', 'IMMEDIATE', 'IMMEDIATELY', 'IMPLEMENTATION',
    'IMPORT', 'IN', 'INCLUDING', 'INCREMENT',
    'INDICATOR', 'INITIAL', 'INITIALLY', 'INNER',
    'INOUT', 'INPUT', 'INSENSITIVE', 'INSERT',
    'INSTANCE', 'INSTANTIABLE', 'INT', 'INTEGER',
    'INTERSECT', 'INTERSECTION', 'INTERVAL', 'INTO',
    'INVOKER', 'IS', 'ISOLATION', 'JAVA',
    'JOIN', 'JSON', 'K', 'KEY',
    'KEY_MEMBER', 'KEY_TYPE', 'LABEL', 'LAG',
    'LANGUAGE', 'LARGE', 'LAST', 'LAST_VALUE',
    'LATERAL', 'LEAD', 'LEADING', 'LEFT',
    'LENGTH', 'LEVEL', 'LIBRARY', 'LIKE',
    'LIKE_REGEX', 'LIMIT', 'LN', 'LOCAL',
    'LOCALTIME', 'LOCALTIMESTAMP', 'LOCATOR', 'LOWER',
    'M', 'MAP', 'MATCH', 'MATCHED',
    'MATCHES', 'MATCH_NUMBER', 'MATCH_RECOGNIZE', 'MAX',
    'MAXVALUE', 'MEASURES', 'MEMBER', 'MERGE',
    'MESSAGE_LENGTH', 'MESSAGE_OCTET_LENGTH', 'MESSAGE_TEXT', 'METHOD',
    'MICROSECOND', 'MILLENNIUM', 'MIN', 'MINUS',
    'MINUTE', 'MINVALUE', 'MOD', 'MODIFIES',
    'MODULE', 'MONTH', 'MORE', 'MULTISET',
    'MUMPS', 'NAME', 'NAMES', 'NATIONAL',
    'NATURAL', 'NCHAR', 'NCLOB', 'NESTING',
    'NEW', 'NEXT', 'NO', 'NONE',
    'NORMALIZE', 'NORMALIZED', 'NOT', 'NTH_VALUE',
    'NTILE', 'NULL', 'NULLABLE', 'NULLIF',
    'NULLS', 'NUMBER', 'NUMERIC', 'OBJECT',
    'OCCURRENCES_REGEX', 'OCTETS', 'OCTET_LENGTH', 'OF',
    'OFFSET', 'OLD', 'OMIT', 'ON',
    'ONE', 'ONLY', 'OPEN', 'OPTION',
    'OPTIONS', 'OR', 'ORDER', 'ORDERING',
    'ORDINALITY', 'OTHERS', 'OUT', 'OUTER',
    'OUTPUT', 'OVER', 'OVERLAPS', 'OVERLAY',
    'OVERRIDING', 'PAD', 'PARAMETER', 'PARAMETER_MODE',
    'PARAMETER_NAME', 'PARAMETER_ORDINAL_POSITION', 'PARAMETER_SPECIFIC_CATALOG', 'PARAMETER_SPECIFIC_NAME',
    'PARAMETER_SPECIFIC_SCHEMA', 'PARTIAL', 'PARTITION', 'PASCAL',
    'PASSTHROUGH', 'PAST', 'PATH', 'PATTERN',
    'PER', 'PERCENT', 'PERCENTILE_CONT', 'PERCENTILE_DISC',
    'PERCENT_RANK', 'PERIOD', 'PERMUTE', 'PLACING',
    'PLAN', 'PLI', 'PORTION', 'POSITION',
    'POSITION_REGEX', 'POWER', 'PRECEDES', 'PRECEDING',
    'PRECISION', 'PREPARE', 'PRESERVE', 'PREV',
    'PRIMARY', 'PRIOR', 'PRIVILEGES', 'PROCEDURE',
    'PUBLIC', 'QUARTER', 'RANGE', 'RANK',
    'READ', 'READS', 'REAL', 'RECURSIVE',
    'REF', 'REFERENCES', 'REFERENCING', 'REGR_AVGX',
    'REGR_AVGY', 'REGR_COUNT', 'REGR_INTERCEPT', 'REGR_R2',
    'REGR_SLOPE', 'REGR_SXX', 'REGR_SXY', 'REGR_SYY',
    'RELATIVE', 'RELEASE', 'REPEATABLE', 'REPLACE',
    'RESET', 'RESTART', 'RESTRICT', 'RESULT',
    'RETURN', 'RETURNED_CARDINALITY', 'RETURNED_LENGTH', 'RETURNED_OCTET_LENGTH',
    'RETURNED_SQLSTATE', 'RETURNS', 'REVOKE', 'RIGHT',
    'ROLE', 'ROLLBACK', 'ROLLUP', 'ROUTINE',
    'ROUTINE_CATALOG', 'ROUTINE_NAME', 'ROUTINE_SCHEMA', 'ROW',
    'ROWS', 'ROW_COUNT', 'ROW_NUMBER', 'RUNNING',
    'SAVEPOINT', 'SCALE', 'SCHEMA', 'SCHEMA_NAME',
    'SCOPE', 'SCOPE_CATALOGS', 'SCOPE_NAME', 'SCOPE_SCHEMA',
    'SCROLL', 'SEARCH', 'SECOND', 'SECTION',
    'SECURITY', 'SEEK', 'SELECT', 'SELF',
    'SENSITIVE', 'SEQUENCE', 'SERIALIZABLE', 'SERVER',
    'SERVER_NAME', 'SESSION', 'SESSION_USER', 'SET',
    'SETS', 'SHOW', 'SIMILAR', 'SIMPLE',
    'SIZE', 'SKIP', 'SMALLINT', 'SOME',
    'SOURCE', 'SPACE', 'SPECIFIC', 'SPECIFICTYPE',
    'SPECIFIC_NAME', 'SQL', 'SQLEXCEPTION', 'SQLSTATE',
    'SQLWARNING', 'SQL_BIGINT', 'SQL_BINARY', 'SQL_BIT',
    'SQL_BLOB', 'SQL_BOOLEAN', 'SQL_CHAR', 'SQL_CLOB',
    'SQL_DATE', 'SQL_DECIMAL', 'SQL_DOUBLE', 'SQL_FLOAT',
    'SQL_INTEGER', 'SQL_INTERVAL_DAY', 'SQL_INTERVAL_DAY_TO_HOUR', 'SQL_INTERVAL_DAY_TO_MINUTE',
    'SQL_INTERVAL_DAY_TO_SECOND', 'SQL_INTERVAL_HOUR', 'SQL_INTERVAL_HOUR_TO_MINUTE', 'SQL_INTERVAL_HOUR_TO_SECOND',
    'SQL_INTERVAL_MINUTE', 'SQL_INTERVAL_MINUTE_TO_SECOND', 'SQL_INTERVAL_MONTH', 'SQL_INTERVAL_SECOND',
    'SQL_INTERVAL_YEAR', 'SQL_INTERVAL_YEAR_TO_MONTH', 'SQL_LONGVARBINARY', 'SQL_LONGVARCHAR',
    'SQL_LONGVARNCHAR', 'SQL_NCHAR', 'SQL_NCLOB', 'SQL_NUMERIC',
    'SQL_NVARCHAR', 'SQL_REAL', 'SQL_SMALLINT', 'SQL_TIME',
    'SQL_TIMESTAMP', 'SQL_TINYINT', 'SQL_TSI_DAY', 'SQL_TSI_FRAC_SECOND',
    'SQL_TSI_HOUR', 'SQL_TSI_MICROSECOND', 'SQL_TSI_MINUTE', 'SQL_TSI_MONTH',
    'SQL_TSI_QUARTER', 'SQL_TSI_SECOND', 'SQL_TSI_WEEK', 'SQL_TSI_YEAR',
    'SQL_VARBINARY', 'SQL_VARCHAR', 'SQRT', 'START',
    'STATE', 'STATEMENT', 'STATIC', 'STDDEV_POP',
    'STDDEV_SAMP', 'STREAM', 'STRUCTURE', 'STYLE',
    'SUBCLASS_ORIGIN', 'SUBMULTISET', 'SUBSET', 'SUBSTITUTE',
    'SUBSTRING', 'SUBSTRING_REGEX', 'SUCCEEDS', 'SUM',
    'SYMMETRIC', 'SYSTEM', 'SYSTEM_TIME', 'SYSTEM_USER',
    'TABLE', 'TABLESAMPLE', 'TABLE_NAME', 'TEMPORARY',
    'THEN', 'TIES', 'TIME', 'TIMESTAMP',
    'TIMESTAMPADD', 'TIMESTAMPDIFF', 'TIMEZONE_HOUR', 'TIMEZONE_MINUTE',
    'TINYINT', 'TO', 'TOP_LEVEL_COUNT', 'TRAILING',
    'TRANSACTION', 'TRANSACTIONS_ACTIVE', 'TRANSACTIONS_COMMITTED', 'TRANSACTIONS_ROLLED_BACK',
    'TRANSFORM', 'TRANSFORMS', 'TRANSLATE', 'TRANSLATE_REGEX',
    'TRANSLATION', 'TREAT', 'TRIGGER', 'TRIGGER_CATALOG',
    'TRIGGER_NAME', 'TRIGGER_SCHEMA', 'TRIM', 'TRIM_ARRAY',
    'TRUE', 'TRUNCATE', 'TYPE', 'UESCAPE',
    'UNBOUNDED', 'UNCOMMITTED', 'UNDER', 'UNION',
    'UNIQUE', 'UNKNOWN', 'UNNAMED', 'UNNEST',
    'UPDATE', 'UPPER', 'UPSERT', 'USAGE',
    'USER', 'USER_DEFINED_TYPE_CATALOG', 'USER_DEFINED_TYPE_CODE', 'USER_DEFINED_TYPE_NAME',
    'USER_DEFINED_TYPE_SCHEMA', 'USING', 'VALUE', 'VALUES',
    'VALUE_OF', 'VARBINARY', 'VARCHAR', 'VARYING',
    'VAR_POP', 'VAR_SAMP', 'VERSION', 'VERSIONING',
    'VIEW', 'WEEK', 'WHEN', 'WHENEVER',
    'WHERE', 'WIDTH_BUCKET', 'WINDOW', 'WITH',
    'WITHIN', 'WITHOUT', 'WORK', 'WRAPPER',
    'WRITE', 'XML', 'YEAR', 'ZONE',
])



#!/usr/bin/env python

from __future__ import division
from __future__ import absolute_import

import sys
import json
import logging
import click

from kylinpy import Kylinpy


def exception_handler(exception_type, exception, traceback):
    print("%s: %s" % (exception_type.__name__, exception))


sys.excepthook = exception_handler  # noqa


@click.group()
@click.option('-h', '--host', required=True, help='kylin/kap host name')
@click.option('-P', '--port', default=7070, help='kylin/kap port, default: 7070')
@click.option('-u', '--username', required=True, help='kylin/kap username')
@click.option('-p', '--password', required=True, help='kylin/kap password')
@click.option('--project', required=True, help='kylin/kap project')
@click.option('--prefix', default='/kylin/api', help='kylin/kap RESTful prefix of url, default: /kylin/api')
@click.option('--debug/--no-debug', default=False, help='show debug infomation')
@click.option('--api2/--api1', default=False, help='API version; default is api1; --api1 used by KYLIN; --api2 used by KAP')
@click.pass_context
def main(ctx, host, port, username, password, project, prefix, debug, api2):
    if debug:
        logging.basicConfig(level=logging.DEBUG)

    _version = 'v2' if api2 else 'v1'
    ctx.obj = Kylinpy(host, username, password, port, project, **{
        'version': _version,
        'prefix': prefix,
        'is_debug': debug
    })


@main.command(help='list all projects')
@click.pass_context
def projects(ctx):
    print(json.dumps(ctx.obj.projects(), indent=4, sort_keys=True))


@main.command(help='sql query')
@click.option('--sql', required=True, help='ANSI-SQL select statement')
@click.pass_context
def query(ctx, sql):
    print(json.dumps(ctx.obj.query(sql), indent=4, sort_keys=True))


@main.command(help='list all users, need admin role, KAP only')
@click.pass_context
def users(ctx):
    print(json.dumps(ctx.obj.users(), indent=4, sort_keys=True))


@main.command(help='get user auth info')
@click.pass_context
def auth(ctx):
    print(json.dumps(ctx.obj.authentication(), indent=4, sort_keys=True))

@main.command(help='show cubes')
@click.pass_context
def cubes(ctx):
    print(json.dumps(ctx.obj.cubes(), indent=4, sort_keys=True))

@main.command(help='get sample sql of cube, KAP only')
@click.option('--name', required=True, help='cube name')
@click.pass_context
def cube_sql(ctx, name):
    print(json.dumps(ctx.obj.cube_sql(name), indent=4, sort_keys=True))


@main.command(help='show cube description')
@click.option('--name', required=True, help='cube name')
@click.pass_context
def cube_desc(ctx, name):
    print(json.dumps(ctx.obj.cube_desc(name), indent=4, sort_keys=True))


@main.command(help='list cube names')
@click.pass_context
def cube_names(ctx):
    print(json.dumps(ctx.obj.get_cube_names(), indent=4, sort_keys=True))


@main.command(help='list cube columns')
@click.option('--name', required=True, help='cube name')
@click.pass_context
def cube_columns(ctx, name):
    print(json.dumps(ctx.obj.get_cube_columns(name), indent=4, sort_keys=True))


@main.command(help='list cube metrics')
@click.option('--name', required=True, help='cube name')
@click.pass_context
def cube_measures(ctx, name):
    print(json.dumps(ctx.obj.get_cube_measures(name), indent=4, sort_keys=True))


@main.command(help='list all table names')
@click.pass_context
def table_names(ctx):
    print(json.dumps(ctx.obj.get_table_names(), indent=4, sort_keys=True))


@main.command(help='list table columns')
@click.option('--name', required=True, help='cube name')
@click.pass_context
def table_columns(ctx, name):
    print(json.dumps(ctx.obj.get_table_columns(name), indent=4, sort_keys=True))


@main.command(help='show model description')
@click.option('--name', required=True, help='model name')
@click.pass_context
def model_desc(ctx, name):
    print(json.dumps(ctx.obj.model_desc(name), indent=4, sort_keys=True))



# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import absolute_import

import re
import json
import base64
import contextlib

from six.moves import urllib

from .errors import (
    KylinUnauthorizedError,
    KylinUserDisabled,
    KylinConnectionError,
    KylinError,
    KylinConfusedResponse,
    KAPOnlyError
)
from .logger import logger
from .utils._compat import as_unicode


class Client(object):
    def __init__(self, scheme, host, port, username, password, **kwargs):
        self.scheme = scheme
        self.host = re.sub('/$', '', host)
        self.port = port
        self.username = username
        self.password = password
        self.version = kwargs.get('version', 'v1')
        self.prefix = re.sub('(^/|/$)', '', kwargs.get('prefix', 'kylin/api'))

    def _prepare_url(self, endpoint, query=None):
        if endpoint.startswith('/'):
            endpoint = endpoint[1:]

        url = '{self.scheme}://{self.host}:{self.port}/{self.prefix}/{endpoint}'.format(
            **locals())

        if query:
            url = '{}?{}'.format(url, urllib.parse.urlencode(query))

        return url

    def _prepare_headers(self, method='GET', session=False):
        headers = {
            'User-Agent': 'Kylin Python Client'
        }

        if not session:
            _auth_str = as_unicode('{}:{}').format(
                self.username, self.password)
            _auth = base64.b64encode(
                _auth_str.encode('utf-8')
            ).decode('ascii')
            headers.update({'Authorization': 'Basic {}'.format(_auth)})
        else:
            # todo session
            pass

        if method in ['POST', 'PUT']:
            headers['Content-Type'] = 'application/json'

        return headers

    def _prepare_body(self, body=None):
        if body:
            body = json.dumps(body).encode('utf-8')

        return body

    def fetch(self, endpoint, method='GET', body=None, params=None):
        try:
            method = method.upper()
            url = self._prepare_url(endpoint, params)
            headers = self._prepare_headers(method)
            if self.version == 'v2':
                headers.update(
                    {'Accept': 'application/vnd.apache.kylin-v2+json'})
            body = self._prepare_body(body)

            req = urllib.request.Request(url, headers=headers)
            req.get_method = lambda: method
            logger.debug("""
==========================[QUERY]===============================
 method: %s \n url: %s \n headers: %s \n body: %s
==========================[QUERY]===============================
            """, method, url, headers, body)

            with contextlib.closing(urllib.request.urlopen(req, body)) as fd:
                try:
                    dumps = json.loads(fd.read().decode("utf-8"))
                except ValueError:
                    raise KylinError('KYLIN JSON object could not decoded')

            return dumps

        except urllib.error.HTTPError as e:
            err = e.read()
            try:
                err = json.loads(err.decode("utf-8"))['msg']
            except(ValueError, AttributeError, KeyError):
                logger.debug(err)
                raise KylinConfusedResponse('Confused Response')

            if e.code == 401 and 'User is disabled' in err:
                raise KylinUserDisabled(err)

            if e.code == 401:
                raise KylinUnauthorizedError(err)

            if e.code == 500:
                raise KylinError(err)

        except urllib.error.URLError as e:
            raise KylinConnectionError("{}".format(str(e.args[0])))


def _is_v2(obj):
    return isinstance(obj, dict) and {'code', 'data', 'msg'} == set(obj.keys())


def compact_response(extract_v2=None, extract_v1=None):
    def fn(request):
        def wrapper(self, *args, **kwargs):
            obj = request(self, *args, **kwargs)
            # todo:
            # if self.is_debug
            #     return obj

            # always return {'data': ENTITY}
            if _is_v2(obj):
                obj = obj['data'].get(
                    extract_v2, None) if extract_v2 else obj['data']
            else:
                obj = obj.get(extract_v1, None) if extract_v1 else obj

            return {'data': obj}
        return wrapper
    return fn


def only_kap_api(fn):
    def wrapper(self, *args, **kwargs):
        if self.client.version == 'v1':
            raise KAPOnlyError
        return fn(self, *args, **kwargs)
    return wrapper


def cache_resp(fn):
    cached = {}

    def wrapper(*args):
        if fn.__name__ not in cached:
            cached[fn.__name__] = fn(*args)
        return cached[fn.__name__]
    return wrapper


class _OriginalAPIMixin(object):
    @compact_response(extract_v1='userDetails')
    def authentication(self):
        return self.client.fetch(endpoint='user/authentication', method='POST')

    @compact_response(extract_v2='projects')
    def projects(self):
        _params = {
            'pageSize': 1000,
            'pageOffset': 0
        }
        return self.client.fetch(endpoint='projects', params=_params)

    @compact_response()
    def query(self, sql, **body):
        _body = {
            'acceptPartial': False,
            'limit': 50000,
            'offset': 0,
            'project': self.project,
            'sql': sql
        }
        _body.update(body)

        return self.client.fetch(endpoint='query', method='POST', body=_body)

    @compact_response()
    def tables_and_columns(self):
        params = {'project': self.project}
        return self.client.fetch(endpoint='tables_and_columns', params=params)

    @compact_response()
    def tables(self, ext=False):
        _params = {
            'project': self.project,
            'ext': ext
        }
        return self.client.fetch(endpoint='tables', params=_params)

    @compact_response(extract_v2='cubes')
    def cubes(self, **params):
        _params = {
            'offset': 0,
            'limit': 50000,
            'projectName': self.project
        }
        _params.update(params)
        return self.client.fetch(endpoint='cubes', params=_params)

    @compact_response()
    @only_kap_api
    def cube_sql(self, cube_name):
        _endpoint = 'cubes/{}/sql'.format(cube_name)
        return self.client.fetch(endpoint=_endpoint)

    @compact_response(extract_v2='cube')
    def cube_desc(self, name):
        if self.client.version == 'v2':
            return self.client.fetch('cube_desc/{}/{}'.format(self.project, name))
        if self.client.version == 'v1':
            return self.client.fetch('cube_desc/{}/desc'.format(name))

    @compact_response(extract_v2="users")
    @only_kap_api
    def users(self):
        _params = {
            'pageSize': 1000,
            'pageOffset': 0
        }
        return self.client.fetch('kap/user/users', params=_params)

    @compact_response(extract_v2="model")
    def model_desc(self, name):
        if self.client.version == 'v2':
            return self.client.fetch('model_desc/{self.project}/{name}'.format(**locals()))
        # v1 did not implement model_desc/<project>/<model_name>
        if self.client.version == 'v1':
            models = self.client.fetch(
                'models', params={'projectName': self.project})
            return [e for e in models if e['name'] == name][0]


class _ExtendedAPIMixin(object):
    def _get_column_datatype(self, column_name, table_name):
        table = [t for t in self.tables()['data'] if t['name'] == table_name][0]
        return [c for c in table['columns'] if c['name'] == column_name][0]['datatype']

    @compact_response()
    def get_table_names(self):
        return [t['table_NAME'] for t in self.tables_and_columns()['data']]

    @compact_response()
    def list_schemas(self):
        table_schemas = [t['table_SCHEM'] for t in self.tables_and_columns()['data']]
        table_schemas = list(set(table_schemas))
        return table_schemas

    @compact_response()
    def get_table_columns(self, table):
        """
        columns list like this:
        [
          {
            u'table': u'CUSTOMER',
            u'column_NAME': u'C_NAME'
            u'datatype': "varchar(256)"
          },
          ...
        ]
        """
        table = [t for t in self.tables_and_columns()['data']
                 if t['table_NAME'] == table]
        columns = table[0]['columns'] if table else []

        return [
            {
                'table_NAME': col['table_NAME'],
                'column_NAME': col['column_NAME'],
                'datatype': self._get_column_datatype(col['column_NAME'], col['table_NAME'])
            }
            for col in columns
        ]

    @compact_response()
    def get_cube_names(self):
        return [cube['name'] for cube in self.cubes()['data'] if cube['status'] == 'READY']

    @compact_response()
    def get_cube_columns(self, cube_name):
        """
        columns list like this:
        [
          {
            u'column': u'C_NAME',
            u'table': u'CUSTOMER',
            u'derived': None,
            u'name': u'C_NAME',
            u'column_NAME': u'C_NAME' <---- tabename_columnname
            u'datatype': "varchar(256)"
          },
          ...
        ]
        """
        cube_desc = self.cube_desc(cube_name)['data']
        dimensions = cube_desc['dimensions']
        model_name = cube_desc['model_name']
        model_desc = self.model_desc(model_name)['data']

        # collect dimension and derived
        collect_dimensions = [dict(dim, **{
            'column_NAME': dim['column'] if dim['derived'] is None else dim['derived'][0]
        }) for dim in dimensions]

        def _get_origin_table(table_name):
            # lookup table name -> fact table name
            origin_table = ([e for e in model_desc['lookups']
                             if e['alias'] == table_name] or [None])[0]
            if origin_table:
                return origin_table['table'].split('.')[1]
            else:
                return table_name

        return [
            dict(dim, **{
                'datatype': self._get_column_datatype(dim['column_NAME'], _get_origin_table(dim['table'])),
                'column_NAME': '_'.join([dim['table'], dim['column_NAME']])
            })
            for dim in collect_dimensions
        ]

    @compact_response()
    def get_cube_measures(self, cube_name):
        cube_measures = self.cube_desc(cube_name)['data']['measures']
        return cube_measures


class Kylinpy(_OriginalAPIMixin, _ExtendedAPIMixin):
    def __init__(self, host, username, password, port=7070, project='default', **kwargs):
        if host.startswith(('http://', 'https://')):
            scheme, host = host.split('://')
        else:
            scheme, host = ('http', host)

        self.client = Client(scheme, host, port, as_unicode(
            username), as_unicode(password), **kwargs)
        self.project = project
        self.is_debug = kwargs.get('is_debug', False)

    def __str__(self):
        c = self.client
        return 'kylin://{c.username}:{c.password}@{c.host}:{c.port}/{self.project}'.format(**locals())



import codecs
import os.path
import re

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='kylinpy',
    version=find_version("kylinpy", "__init__.py"),
    author='Yongjie Zhao',
    author_email='yongjie.zhao@kyligence.io',
    maintainer='Yongjie Zhao',
    maintainer_email='yongjie.zhao@kyligence.io',
    packages=['kylinpy', 'kylinpy.utils'],
    url='https://github.com/Kyligence/kylinpy',
    license='MIT License',
    description='Apache Kylin Python Client Library',
    long_description=open('README.rst').read(),
    install_requires=['six>=1.0.0', 'click>=6.7'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'coverage', 'mock'],
    extras_require={
        'sqlalchemy': ['sqlalchemy>=1.1.0'],
    },
    keywords=['kylin', 'kap', 'cli', 'sqlalchemy dialect'],
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    entry_points={
        "sqlalchemy.dialects": ["kylin=kylinpy.sqla_dialect:KylinDialect"],
        'console_scripts': ['kylinpy=kylinpy.cli:main'],
    }
)



import sys

PY3 = sys.version_info[0] == 3

if PY3:
    def as_unicode(s):
        if isinstance(s, bytes):
            return s.decode('utf-8')
        return str(s)

else:
    def as_unicode(s):
        if isinstance(s, str):
            return s.decode('utf-8')
        return unicode(s)



from __future__ import absolute_import

import re
import itertools

from sqlalchemy import pool
from sqlalchemy.sql import compiler
from sqlalchemy.engine import default

from .kylindb import KylinDB
from .utils.sqla_types import kylin_to_sqla
from .utils.keywords import CALCITE_KEYWORDS

SUPERSET_KEYWORDS = set([
    '__timestamp',
])


class KylinIdentifierPreparer(compiler.IdentifierPreparer):
    compiler.IdentifierPreparer.reserved_words = set(
        itertools.chain(*[[e.lower(), e] for e in CALCITE_KEYWORDS])
    )
    compiler.IdentifierPreparer.reserved_words.update(SUPERSET_KEYWORDS)

    def __init__(self, dialect, initial_quote='"',
                 final_quote=None, escape_quote='"', omit_schema=True):
        super(KylinIdentifierPreparer, self).__init__(
            dialect, initial_quote, final_quote, escape_quote, omit_schema
        )


class KylinSQLCompiler(compiler.SQLCompiler):

    def __init__(self, *args, **kwargs):
        super(KylinSQLCompiler, self).__init__(*args, **kwargs)

    def _compose_select_body(self, text, select, inner_columns, froms, byfrom, kwargs):
        text = super(KylinSQLCompiler, self)._compose_select_body(
            text, select, inner_columns, froms, byfrom, kwargs)
        # todo
        text = re.sub(' \d{2}:\d{2}:\d{2}', '', text)
        return text


class KylinDialect(default.DefaultDialect):
    name = 'kylin'
    driver = 'kylin'

    statement_compiler = KylinSQLCompiler
    preparer = KylinIdentifierPreparer

    preexecute_pk_sequences = True
    supports_pk_autoincrement = True
    supports_sequences = True
    sequences_optional = True
    supports_native_decimal = True
    supports_default_values = True
    supports_native_boolean = True
    poolclass = pool.SingletonThreadPool
    supports_unicode_statements = True

    default_paramstyle = 'pyformat'

    def __init__(self, *args, **kwargs):
        super(KylinDialect, self).__init__(*args, **kwargs)

    @classmethod
    def dbapi(cls):
        return KylinDB

    def initialize(self, connection):
        self.server_version_info = None
        self.default_schema_name = None
        self.default_isolation_level = None
        self.returns_unicode_strings = True

    def create_connect_args(self, url):
        args = {
            'username': url.username,
            'password': url.password,
            'host': url.host,
            'port': url.port,
            'project': re.sub('/$', '', url.database or 'default'),
            'version': url.query.get('version'),
            'prefix': url.query.get('prefix'),
        }
        return [], dict([e for e in args.items() if e[1]])

    def get_table_names(self, connection, schema=None, **kw):
        conn = connection.connect()
        return conn.connection.connection.get_table_names().get('data')

    def get_schema_names(self, connection, schema=None, **kw):
        conn = connection.connect()
        return conn.connection.connection.list_schemas().get('data')

    def has_table(self, connection, table_name, schema=None):
        return table_name in self.get_table_names(connection, schema)

    def has_sequence(self, connection, sequence_name, schema=None):
        return False

    def get_columns(self, connection, table_name, schema=None, **kw):
        conn = connection.connect()
        columns = conn.connection.connection.get_table_columns(
            table_name).get('data')
        return [{
            'name': col['column_NAME'],
            'type': kylin_to_sqla(col['datatype'])
        } for col in columns]

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        return []

    def get_indexes(self, connection, table_name, schema=None, **kw):
        return []

    def get_view_names(self, connection, schema=None, **kw):
        return []

    def get_pk_constraint(self, conn, table_name, schema=None, **kw):
        return {}

    def get_unique_constraints(self, connection, table_name, schema=None, **kw):
        return []
